@@ -8,11 +8,13 @@
 /*
 /*	int	qmgr_message_count;
 /*	int	qmgr_recipient_count;
+/*	int	qmgr_vrfy_pend_count;
 /*
-/*	QMGR_MESSAGE *qmgr_message_alloc(class, name, qflags)
+/*	QMGR_MESSAGE *qmgr_message_alloc(class, name, qflags, mode)
 /*	const char *class;
 /*	const char *name;
 /*	int	qflags;
+/*	mode_t	mode;
 /*
 /*	QMGR_MESSAGE *qmgr_message_realloc(message)
 /*	QMGR_MESSAGE *message;
@@ -37,6 +39,13 @@
 /*	of in-core recipient structures (i.e. the sum of all recipients
 /*	in all in-core message structures).
 /*
+/*	qmgr_vrfy_pend_count is a global counter for the total
+/*	number of in-core message structures that are associated
+/*	with an address verification request. Requests that exceed
+/*	the address_verify_pending_limit are deferred immediately.
+/*	This is a backup mechanism for a more refined enforcement
+/*	mechanism in the verify(8) daemon.
+/*
 /*	qmgr_message_alloc() creates an in-core message structure
 /*	with sender and recipient information taken from the named queue
 /*	file. A null result means the queue file could not be read or
@@ -49,6 +58,7 @@
 /*	run through the resolver, and are assigned to destination
 /*	queues. Recipients that cannot be assigned are deferred or
 /*	bounced. Mail that has bounced twice is silently absorbed.
+/*	A non-zero mode means change the queue file permissions.
 /*
 /*	qmgr_message_realloc() resumes reading recipients from the queue
 /*	file, and updates the recipient list and \fIrcpt_offset\fR message
@@ -79,6 +89,11 @@
 /*	IBM T.J. Watson Research
 /*	P.O. Box 704
 /*	Yorktown Heights, NY 10598, USA
+/*
+/*	Wietse Venema
+/*	Google, Inc.
+/*	111 8th Avenue
+/*	New York, NY 10011, USA
 /*--*/
 
 /* System library. */
@@ -93,10 +108,6 @@
 #include <string.h>
 #include <ctype.h>
 
-#ifdef STRCASECMP_IN_STRINGS_H
-#include <strings.h>
-#endif
-
 /* Utility library. */
 
 #include <msg.h>
@@ -125,7 +136,7 @@
 #include <qmgr_user.h>
 #include <split_addr.h>
 #include <dsn_mask.h>
-#include <dsn_attr_map.h>
+#include <rec_attr_map.h>
 
 /* Client stubs. */
 
@@ -138,6 +149,7 @@
 
 int     qmgr_message_count;
 int     qmgr_recipient_count;
+int     qmgr_vrfy_pend_count;
 
 /* qmgr_message_create - create in-core message structure */
 
@@ -157,6 +169,7 @@ static QMGR_MESSAGE *qmgr_message_create(const char *queue_name,
     message->refcount = 0;
     message->single_rcpt = 0;
     message->arrival_time.tv_sec = message->arrival_time.tv_usec = 0;
+    message->create_time = 0;
     GETTIMEOFDAY(&message->active_time);
     message->data_offset = 0;
     message->queue_id = mystrdup(queue_id);
@@ -165,21 +178,25 @@ static QMGR_MESSAGE *qmgr_message_create(const char *queue_name,
     message->sender = 0;
     message->dsn_envid = 0;
     message->dsn_ret = 0;
+    message->smtputf8 = 0;
     message->filter_xport = 0;
     message->inspect_xport = 0;
     message->redirect_addr = 0;
     message->data_size = 0;
+    message->cont_length = 0;
     message->warn_offset = 0;
     message->warn_time = 0;
     message->rcpt_offset = 0;
     message->verp_delims = 0;
     message->client_name = 0;
     message->client_addr = 0;
+    message->client_port = 0;
     message->client_proto = 0;
     message->client_helo = 0;
     message->sasl_method = 0;
     message->sasl_username = 0;
     message->sasl_sender = 0;
+    message->log_ident = 0;
     message->rewrite_context = 0;
     recipient_list_init(&message->rcpt_list, RCPT_LIST_INIT_QUEUE);
     return (message);
@@ -306,6 +323,7 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
     int     dsn_notify = 0;
     char   *dsn_orcpt = 0;
     int     n;
+    int     have_log_client_attr = 0;
 
     /*
      * Initialize. No early returns or we have a memory leak.
@@ -360,10 +378,16 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 		msg_fatal("seek file %s: %m", VSTREAM_PATH(message->fp));
 	    curr_offset += message->data_size;
 	}
-	rec_type = rec_get(message->fp, buf, 0);
+	rec_type = rec_get_raw(message->fp, buf, 0, REC_FLAG_NONE);
 	start = vstring_str(buf);
 	if (msg_verbose > 1)
 	    msg_info("record %c %s", rec_type, start);
+	if (rec_type == REC_TYPE_PTR) {
+	    if ((rec_type = rec_goto(message->fp, start)) == REC_TYPE_ERROR)
+		break;
+	    /* Need to update curr_offset after pointer jump. */
+	    continue;
+	}
 	if (rec_type <= 0) {
 	    msg_warn("%s: message rejected: missing end record",
 		     message->queue_id);
@@ -387,7 +411,7 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 		rec_type = REC_TYPE_ERROR;
 		break;
 	    }
-	    if ((n = dsn_attr_map(name)) != 0) {
+	    if ((n = rec_attr_map(name)) != 0) {
 		start = value;
 		rec_type = n;
 	    }
@@ -439,7 +463,7 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 	    }
 	    continue;
 	}
-	if (rec_type == REC_TYPE_DONE) {
+	if (rec_type == REC_TYPE_DONE || rec_type == REC_TYPE_DRCP) {
 	    if (message->rcpt_offset == 0) {
 		if (dsn_orcpt) {
 		    myfree(dsn_orcpt);
@@ -503,9 +527,11 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 	    continue;
 	if (rec_type == REC_TYPE_SIZE) {
 	    if (message->data_offset == 0) {
-		if ((count = sscanf(start, "%ld %ld %d %d",
+		if ((count = sscanf(start, "%ld %ld %d %d %ld %d",
 				 &message->data_size, &message->data_offset,
-				    &nrcpt, &message->rflags)) >= 3) {
+				    &nrcpt, &message->rflags,
+				    &message->cont_length,
+				    &message->smtputf8)) >= 3) {
 		    /* Postfix >= 1.0 (a.k.a. 20010228). */
 		    if (message->data_offset <= 0 || message->data_size <= 0) {
 			msg_warn("%s: invalid size record: %.100s",
@@ -530,6 +556,15 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 		    break;
 		}
 	    }
+	    /* Postfix < 2.4 compatibility. */
+	    if (message->cont_length == 0) {
+		message->cont_length = message->data_size;
+	    } else if (message->cont_length < 0) {
+		msg_warn("%s: invalid size record: %.100s",
+			 message->queue_id, start);
+		rec_type = REC_TYPE_ERROR;
+		break;
+	    }
 	    continue;
 	}
 	if (rec_type == REC_TYPE_TIME) {
@@ -537,6 +572,11 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 		REC_TYPE_TIME_SCAN(start, message->arrival_time);
 	    continue;
 	}
+	if (rec_type == REC_TYPE_CTIME) {
+	    if (message->create_time == 0)
+		message->create_time = atol(start);
+	    continue;
+	}
 	if (rec_type == REC_TYPE_FILT) {
 	    if (message->filter_xport != 0)
 		myfree(message->filter_xport);
@@ -559,23 +599,24 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 	    if (message->sender == 0) {
 		message->sender = mystrdup(start);
 		opened(message->queue_id, message->sender,
-		       message->data_size, nrcpt,
+		       message->cont_length, nrcpt,
 		       "queue %s", message->queue_name);
 	    }
 	    continue;
 	}
 	if (rec_type == REC_TYPE_DSN_ENVID) {
-	    if (message->dsn_envid == 0)
-		message->dsn_envid = mystrdup(start);
+	    /* Allow Milter override. */
+	    if (message->dsn_envid != 0)
+		myfree(message->dsn_envid);
+	    message->dsn_envid = mystrdup(start);
 	}
 	if (rec_type == REC_TYPE_DSN_RET) {
-	    if (message->dsn_ret == 0) {
-		if (!alldig(start) || (n = atoi(start)) == 0 || !DSN_RET_OK(n))
-		    msg_warn("%s: ignoring malformed DSN RET flags in queue file record:%.100s",
-			     message->queue_id, start);
-		else
-		    message->dsn_ret = n;
-	    }
+	    /* Allow Milter override. */
+	    if (!alldig(start) || (n = atoi(start)) == 0 || !DSN_RET_OK(n))
+		msg_warn("%s: ignoring malformed DSN RET flags in queue file record:%.100s",
+			 message->queue_id, start);
+	    else
+		message->dsn_ret = n;
 	}
 	if (rec_type == REC_TYPE_ATTR) {
 	    /* Allow extra segment to override envelope segment info. */
@@ -584,23 +625,63 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 		    myfree(message->encoding);
 		message->encoding = mystrdup(value);
 	    }
+
+	    /*
+	     * Backwards compatibility. Before Postfix 2.3, the logging
+	     * attributes were called client_name, etc. Now they are called
+	     * log_client_name. etc., and client_name is used for the actual
+	     * client information. To support old queue files, we accept both
+	     * names for the purpose of logging; the new name overrides the
+	     * old one.
+	     * 
+	     * XXX Do not use the "legacy" client_name etc. attribute values for
+	     * initializing the logging attributes, when this file already
+	     * contains the "modern" log_client_name etc. logging attributes.
+	     * Otherwise, logging attributes that are not present in the
+	     * queue file would be set with information from the real client.
+	     */
+	    else if (strcmp(name, MAIL_ATTR_ACT_CLIENT_NAME) == 0) {
+		if (have_log_client_attr == 0 && message->client_name == 0)
+		    message->client_name = mystrdup(value);
+	    } else if (strcmp(name, MAIL_ATTR_ACT_CLIENT_ADDR) == 0) {
+		if (have_log_client_attr == 0 && message->client_addr == 0)
+		    message->client_addr = mystrdup(value);
+	    } else if (strcmp(name, MAIL_ATTR_ACT_CLIENT_PORT) == 0) {
+		if (have_log_client_attr == 0 && message->client_port == 0)
+		    message->client_port = mystrdup(value);
+	    } else if (strcmp(name, MAIL_ATTR_ACT_PROTO_NAME) == 0) {
+		if (have_log_client_attr == 0 && message->client_proto == 0)
+		    message->client_proto = mystrdup(value);
+	    } else if (strcmp(name, MAIL_ATTR_ACT_HELO_NAME) == 0) {
+		if (have_log_client_attr == 0 && message->client_helo == 0)
+		    message->client_helo = mystrdup(value);
+	    }
 	    /* Original client attributes. */
-	    else if (strcmp(name, MAIL_ATTR_CLIENT_NAME) == 0) {
+	    else if (strcmp(name, MAIL_ATTR_LOG_CLIENT_NAME) == 0) {
 		if (message->client_name != 0)
 		    myfree(message->client_name);
 		message->client_name = mystrdup(value);
-	    } else if (strcmp(name, MAIL_ATTR_CLIENT_ADDR) == 0) {
+		have_log_client_attr = 1;
+	    } else if (strcmp(name, MAIL_ATTR_LOG_CLIENT_ADDR) == 0) {
 		if (message->client_addr != 0)
 		    myfree(message->client_addr);
 		message->client_addr = mystrdup(value);
-	    } else if (strcmp(name, MAIL_ATTR_PROTO_NAME) == 0) {
+		have_log_client_attr = 1;
+	    } else if (strcmp(name, MAIL_ATTR_LOG_CLIENT_PORT) == 0) {
+		if (message->client_port != 0)
+		    myfree(message->client_port);
+		message->client_port = mystrdup(value);
+		have_log_client_attr = 1;
+	    } else if (strcmp(name, MAIL_ATTR_LOG_PROTO_NAME) == 0) {
 		if (message->client_proto != 0)
 		    myfree(message->client_proto);
 		message->client_proto = mystrdup(value);
-	    } else if (strcmp(name, MAIL_ATTR_HELO_NAME) == 0) {
+		have_log_client_attr = 1;
+	    } else if (strcmp(name, MAIL_ATTR_LOG_HELO_NAME) == 0) {
 		if (message->client_helo != 0)
 		    myfree(message->client_helo);
 		message->client_helo = mystrdup(value);
+		have_log_client_attr = 1;
 	    } else if (strcmp(name, MAIL_ATTR_SASL_METHOD) == 0) {
 		if (message->sasl_method == 0)
 		    message->sasl_method = mystrdup(value);
@@ -619,6 +700,12 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 		else
 		    msg_warn("%s: ignoring multiple %s attribute: %s",
 			   message->queue_id, MAIL_ATTR_SASL_SENDER, value);
+	    } else if (strcmp(name, MAIL_ATTR_LOG_IDENT) == 0) {
+		if (message->log_ident == 0)
+		    message->log_ident = mystrdup(value);
+		else
+		    msg_warn("%s: ignoring multiple %s attribute: %s",
+			     message->queue_id, MAIL_ATTR_LOG_IDENT, value);
 	    } else if (strcmp(name, MAIL_ATTR_RWR_CONTEXT) == 0) {
 		if (message->rewrite_context == 0)
 		    message->rewrite_context = mystrdup(value);
@@ -633,11 +720,15 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 	     * after the logfile is deleted.
 	     */
 	    else if (strcmp(name, MAIL_ATTR_TRACE_FLAGS) == 0) {
-		message->tflags = DEL_REQ_TRACE_FLAGS(atoi(value));
-		if (message->tflags == DEL_REQ_FLAG_RECORD)
-		    message->tflags_offset = curr_offset;
-		else
-		    message->tflags_offset = 0;
+		if (message->tflags == 0) {
+		    message->tflags = DEL_REQ_TRACE_FLAGS(atoi(value));
+		    if (message->tflags == DEL_REQ_FLAG_RECORD)
+			message->tflags_offset = curr_offset;
+		    else
+			message->tflags_offset = 0;
+		    if ((message->tflags & DEL_REQ_FLAG_MTA_VRFY) != 0)
+			qmgr_vrfy_pend_count++;
+		}
 	    }
 	    continue;
 	}
@@ -657,6 +748,9 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 		    msg_warn("%s: ignoring bad VERP request: \"%.100s\"",
 			     message->queue_id, start);
 		} else {
+		    if (msg_verbose)
+			msg_info("%s: enabling VERP for sender \"%.100s\"",
+				 message->queue_id, message->sender);
 		    message->single_rcpt = 1;
 		    message->verp_delims = mystrdup(start);
 		}
@@ -672,7 +766,7 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 	if (rec_type > 0)
 	    msg_warn("%s: ignoring out-of-order DSN original recipient <%.200s>",
 		     message->queue_id, dsn_orcpt);
-	myfree(orig_rcpt);
+	myfree(dsn_orcpt);
     }
     if (orig_rcpt != 0) {
 	if (rec_type > 0)
@@ -681,6 +775,16 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 	myfree(orig_rcpt);
     }
 
+    /*
+     * After sending a "delayed" warning, request sender notification when
+     * message delivery is completed. While "mail delayed" notifications are
+     * bad enough because they multiply the amount of email traffic, "delay
+     * cleared" notifications are even worse because they come in a sudden
+     * burst when the queue drains after a network outage.
+     */
+    if (var_dsn_delay_cleared && message->warn_time < 0)
+	message->tflags |= DEL_REQ_FLAG_REC_DLY_SENT;
+
     /*
      * Avoid clumsiness elsewhere in the program. When sending data across an
      * IPC channel, sending an empty string is more convenient than sending a
@@ -694,6 +798,8 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 	message->client_name = mystrdup("");
     if (message->client_addr == 0)
 	message->client_addr = mystrdup("");
+    if (message->client_port == 0)
+	message->client_port = mystrdup("");
     if (message->client_proto == 0)
 	message->client_proto = mystrdup("");
     if (message->client_helo == 0)
@@ -704,8 +810,13 @@ static int qmgr_message_read(QMGR_MESSAGE *message)
 	message->sasl_username = mystrdup("");
     if (message->sasl_sender == 0)
 	message->sasl_sender = mystrdup("");
+    if (message->log_ident == 0)
+	message->log_ident = mystrdup("");
     if (message->rewrite_context == 0)
 	message->rewrite_context = mystrdup(MAIL_ATTR_RWR_LOCAL);
+    /* Postfix < 2.3 compatibility. */
+    if (message->create_time == 0)
+	message->create_time = message->arrival_time.tv_sec;
 
     /*
      * Clean up.
@@ -740,13 +851,13 @@ void    qmgr_message_update_warn(QMGR_MESSAGE *message)
 {
 
     /*
-     * XXX eventually this should let us schedule multiple warnings, right
-     * now it just allows for one.
+     * After the "mail delayed" warning, optionally send a "delay cleared"
+     * notification.
      */
     if (qmgr_message_open(message)
 	|| vstream_fseek(message->fp, message->warn_offset, SEEK_SET) < 0
 	|| rec_fprintf(message->fp, REC_TYPE_WARN, REC_TYPE_WARN_FORMAT,
-		       REC_TYPE_WARN_ARG(0)) < 0
+		       REC_TYPE_WARN_ARG(-1)) < 0
 	|| vstream_fflush(message->fp))
 	msg_fatal("update queue file %s: %m", VSTREAM_PATH(message->fp));
     qmgr_message_close(message);
@@ -815,20 +926,20 @@ static int qmgr_message_sort_compare(const void *p1, const void *p2)
     if (at1 != 0 && at2 == 0)
 	return (-1);
     if (at1 != 0 && at2 != 0
-	&& (result = strcasecmp(at1, at2)) != 0)
+	&& (result = strcasecmp_utf8(at1, at2)) != 0)
 	return (result);
 
     /*
      * Compare recipient address.
      */
-    return (strcasecmp(rcpt1->address, rcpt2->address));
+    return (strcasecmp_utf8(rcpt1->address, rcpt2->address));
 }
 
 /* qmgr_message_sort - sort message recipient addresses by domain */
 
 static void qmgr_message_sort(QMGR_MESSAGE *message)
 {
-    qsort((char *) message->rcpt_list.info, message->rcpt_list.len,
+    qsort((void *) message->rcpt_list.info, message->rcpt_list.len,
 	  sizeof(message->rcpt_list.info[0]), qmgr_message_sort_compare);
     if (msg_verbose) {
 	RECIPIENT_LIST list = message->rcpt_list;
@@ -846,24 +957,24 @@ static void qmgr_message_sort(QMGR_MESSAGE *message)
 static int qmgr_resolve_one(QMGR_MESSAGE *message, RECIPIENT *recipient,
 			            const char *addr, RESOLVE_REPLY *reply)
 {
-    DSN     dsn;
+#define QMGR_REDIRECT(rp, tp, np) do { \
+	(rp)->flags = 0; \
+	vstring_strcpy((rp)->transport, (tp)); \
+	vstring_strcpy((rp)->nexthop, (np)); \
+    } while (0)
 
     if ((message->tflags & DEL_REQ_FLAG_MTA_VRFY) == 0)
 	resolve_clnt_query_from(message->sender, addr, reply);
     else
 	resolve_clnt_verify_from(message->sender, addr, reply);
     if (reply->flags & RESOLVE_FLAG_FAIL) {
-	qmgr_defer_recipient(message, recipient,
-			     DSN_SMTP(&dsn, "4.3.0",
-				      "451 address resolver failure",
-				      "address resolver failure"));
-	return (-1);
+	QMGR_REDIRECT(reply, MAIL_SERVICE_RETRY,
+		      "4.3.0 address resolver failure");
+	return (0);
     } else if (reply->flags & RESOLVE_FLAG_ERROR) {
-	qmgr_bounce_recipient(message, recipient,
-			      DSN_SMTP(&dsn, "5.1.3",
-				       "553 bad address syntax",
-				       "bad address syntax"));
-	return (-1);
+	QMGR_REDIRECT(reply, MAIL_SERVICE_ERROR,
+		      "5.1.3 bad address syntax");
+	return (0);
     } else {
 	return (0);
     }
@@ -887,6 +998,7 @@ static void qmgr_message_resolve(QMGR_MESSAGE *message)
     int     status;
     DSN     dsn;
     MSG_STATS stats;
+    DSN    *saved_dsn;
 
 #define STREQ(x,y)	(strcmp(x,y) == 0)
 #define STR		vstring_str
@@ -897,15 +1009,16 @@ static void qmgr_message_resolve(QMGR_MESSAGE *message)
     for (recipient = list.info; recipient < list.info + list.len; recipient++) {
 
 	/*
-	 * Redirect overrides all else. But only once (per batch of
-	 * recipients). For consistency with the remainder of Postfix,
-	 * rewrite the address to canonical form before resolving it.
+	 * Redirect overrides all else. But only once (per entire message).
+	 * For consistency with the remainder of Postfix, rewrite the address
+	 * to canonical form before resolving it.
 	 */
 	if (message->redirect_addr) {
 	    if (recipient > list.info) {
 		recipient->u.queue = 0;
 		continue;
 	    }
+	    message->rcpt_offset = 0;
 	    rewrite_clnt_internal(REWRITE_CANON, message->redirect_addr,
 				  reply.recipient);
 	    RECIPIENT_UPDATE(recipient->address, STR(reply.recipient));
@@ -918,12 +1031,25 @@ static void qmgr_message_resolve(QMGR_MESSAGE *message)
 
 	/*
 	 * Content filtering overrides the address resolver.
+	 * 
+	 * XXX Bypass content_filter inspection for user-generated probes
+	 * (sendmail -bv). MTA-generated probes never have the "please filter
+	 * me" bits turned on, but we handle them here anyway for the sake of
+	 * future proofing.
 	 */
-	else if (message->filter_xport) {
+#define FILTER_WITHOUT_NEXTHOP(filter, next) \
+	(((next) = split_at((filter), ':')) == 0 || *(next) == 0)
+
+#define RCPT_WITHOUT_DOMAIN(rcpt, next) \
+	((next = strrchr(rcpt, '@')) == 0 || *++(next) == 0)
+
+	else if (message->filter_xport
+		 && (message->tflags & DEL_REQ_TRACE_ONLY_MASK) == 0) {
 	    reply.flags = 0;
 	    vstring_strcpy(reply.transport, message->filter_xport);
-	    if ((nexthop = split_at(STR(reply.transport), ':')) == 0
-		|| *nexthop == 0)
+	    if (FILTER_WITHOUT_NEXTHOP(STR(reply.transport), nexthop)
+		&& *(nexthop = var_def_filter_nexthop) == 0
+		&& RCPT_WITHOUT_DOMAIN(recipient->address, nexthop))
 		nexthop = var_myhostname;
 	    vstring_strcpy(reply.nexthop, nexthop);
 	    vstring_strcpy(reply.recipient, recipient->address);
@@ -947,30 +1073,23 @@ static void qmgr_message_resolve(QMGR_MESSAGE *message)
 	 * the queue manager process does not help.
 	 */
 	if (recipient->address[0] == 0) {
-	    qmgr_bounce_recipient(message, recipient,
-				  DSN_SMTP(&dsn, "5.1.3",
-					   "553 null recipient address",
-					   "null recipient address"));
-	    continue;
+	    QMGR_REDIRECT(&reply, MAIL_SERVICE_ERROR,
+			  "5.1.3 null recipient address");
 	}
 
 	/*
-	 * Bounce recipient addresses that start with `-'. External commands
-	 * may misinterpret such addresses as command-line options.
-	 * 
-	 * In theory I could say people should always carefully set up their
-	 * master.cf pipe mailer entries with `--' before the first
-	 * non-option argument, but mistakes will happen regardless.
-	 * 
-	 * Therefore the protection is put in place here, in the queue manager,
-	 * where it cannot be bypassed.
+	 * Redirect a force-expired message without defer log to the retry
+	 * service, so that its defer log will contain an appropriate reason.
+	 * Do not redirect such a message to the error service, because if
+	 * that request fails, a defer log would be created with reason
+	 * "bounce or trace service failure" which would make no sense. Note
+	 * that if the bounce service fails to create a defer log, the
+	 * message will be returned as undeliverable anyway, because it is
+	 * expired.
 	 */
-	if (var_allow_min_user == 0 && recipient->address[0] == '-') {
-	    qmgr_bounce_recipient(message, recipient,
-				  DSN_SMTP(&dsn, "5.1.3",
-					   "553 bad address syntax",
-					   "bad address syntax"));
-	    continue;
+	if ((message->qflags & QMGR_FORCE_EXPIRE) != 0) {
+	    QMGR_REDIRECT(&reply, MAIL_SERVICE_RETRY,
+			  "4.7.0 message is administratively expired");
 	}
 
 	/*
@@ -986,8 +1105,8 @@ static void qmgr_message_resolve(QMGR_MESSAGE *message)
 	    at = strrchr(STR(reply.recipient), '@');
 	    len = (at ? (at - STR(reply.recipient))
 		   : strlen(STR(reply.recipient)));
-	    if (strncasecmp(STR(reply.recipient), var_double_bounce_sender,
-			    len) == 0
+	    if (strncasecmp_utf8(STR(reply.recipient),
+				 var_double_bounce_sender, len) == 0
 		&& !var_double_bounce_sender[len]) {
 		status = sent(message->tflags, message->queue_id,
 			      QMGR_MSG_STATS(&stats, message), recipient,
@@ -995,8 +1114,11 @@ static void qmgr_message_resolve(QMGR_MESSAGE *message)
 			"undeliverable postmaster notification discarded"));
 		if (status == 0) {
 		    deliver_completed(message->fp, recipient->offset);
+#if 0
+		    /* It's the default verification probe sender address. */
 		    msg_warn("%s: undeliverable postmaster notification discarded",
 			     message->queue_id);
+#endif
 		} else
 		    message->flags |= status;
 		continue;
@@ -1007,21 +1129,26 @@ static void qmgr_message_resolve(QMGR_MESSAGE *message)
 	 * Optionally defer deliveries over specific transports, unless the
 	 * restriction is lifted temporarily.
 	 */
-	if (*var_defer_xports && (message->qflags & QMGR_FLUSH_DEAD) == 0) {
+	if (*var_defer_xports && (message->qflags & QMGR_FLUSH_DFXP) == 0) {
 	    if (defer_xport_argv == 0)
-		defer_xport_argv = argv_split(var_defer_xports, " \t\r\n,");
+		defer_xport_argv = argv_split(var_defer_xports, CHARS_COMMA_SP);
 	    for (cpp = defer_xport_argv->argv; *cpp; cpp++)
 		if (strcmp(*cpp, STR(reply.transport)) == 0)
 		    break;
 	    if (*cpp) {
-		qmgr_defer_recipient(message, recipient,
-				     DSN_SMTP(&dsn, "4.3.2",
-					      "450 delivery suspended",
-					      "deferred transport"));
-		continue;
+		QMGR_REDIRECT(&reply, MAIL_SERVICE_RETRY,
+			      "4.3.2 deferred transport");
 	    }
 	}
 
+	/*
+	 * Safety: defer excess address verification requests.
+	 */
+	if ((message->tflags & DEL_REQ_FLAG_MTA_VRFY) != 0
+	    && qmgr_vrfy_pend_count > var_vrfy_pend_limit)
+	    QMGR_REDIRECT(&reply, MAIL_SERVICE_RETRY,
+			  "4.3.2 Too many address verification requests");
+
 	/*
 	 * Look up or instantiate the proper transport.
 	 */
@@ -1031,12 +1158,28 @@ static void qmgr_message_resolve(QMGR_MESSAGE *message)
 	    queue = 0;
 	}
 
+	/*
+	 * This message is being flushed. If need-be unthrottle the
+	 * transport.
+	 */
+	if ((message->qflags & QMGR_FLUSH_EACH) != 0
+	    && QMGR_TRANSPORT_THROTTLED(transport))
+	    qmgr_transport_unthrottle(transport);
+
 	/*
 	 * This transport is dead. Defer delivery to this recipient.
 	 */
-	if ((transport->flags & QMGR_TRANSPORT_STAT_DEAD) != 0) {
-	    qmgr_defer_recipient(message, recipient, transport->dsn);
-	    continue;
+	if (QMGR_TRANSPORT_THROTTLED(transport)) {
+	    saved_dsn = transport->dsn;
+	    if ((transport = qmgr_error_transport(MAIL_SERVICE_RETRY)) != 0) {
+		nexthop = qmgr_error_nexthop(saved_dsn);
+		vstring_strcpy(reply.nexthop, nexthop);
+		myfree(nexthop);
+		queue = 0;
+	    } else {
+		qmgr_defer_recipient(message, recipient, saved_dsn);
+		continue;
+	    }
 	}
 
 	/*
@@ -1074,6 +1217,7 @@ static void qmgr_message_resolve(QMGR_MESSAGE *message)
 	 */
 	vstring_strcpy(queue_name, STR(reply.nexthop));
 	if (strcmp(transport->name, MAIL_SERVICE_ERROR) != 0
+	    && strcmp(transport->name, MAIL_SERVICE_RETRY) != 0
 	    && transport->recipient_limit == 1) {
 	    /* Copy the recipient localpart. */
 	    at = strrchr(STR(reply.recipient), '@');
@@ -1081,7 +1225,7 @@ static void qmgr_message_resolve(QMGR_MESSAGE *message)
 		   : strlen(STR(reply.recipient)));
 	    vstring_strncpy(queue_name, STR(reply.recipient), len);
 	    /* Remove the address extension from the recipient localpart. */
-	    if (*var_rcpt_delim && split_addr(STR(queue_name), *var_rcpt_delim))
+	    if (*var_rcpt_delim && split_addr(STR(queue_name), var_rcpt_delim))
 		vstring_truncate(queue_name, strlen(STR(queue_name)));
 	    /* Assume the recipient domain is equivalent to nexthop. */
 	    vstring_sprintf_append(queue_name, "@%s", STR(reply.nexthop));
@@ -1098,12 +1242,22 @@ static void qmgr_message_resolve(QMGR_MESSAGE *message)
 					  STR(reply.nexthop));
 	}
 
+	/*
+	 * This message is being flushed. If need-be unthrottle the queue.
+	 */
+	if ((message->qflags & QMGR_FLUSH_EACH) != 0
+	    && QMGR_QUEUE_THROTTLED(queue))
+	    qmgr_queue_unthrottle(queue);
+
 	/*
 	 * This queue is dead. Defer delivery to this recipient.
 	 */
-	if (queue->window == 0) {
-	    qmgr_defer_recipient(message, recipient, queue->dsn);
-	    continue;
+	if (QMGR_QUEUE_THROTTLED(queue)) {
+	    saved_dsn = queue->dsn;
+	    if ((queue = qmgr_error_queue(MAIL_SERVICE_RETRY, saved_dsn)) == 0) {
+		qmgr_defer_recipient(message, recipient, saved_dsn);
+		continue;
+	    }
 	}
 
 	/*
@@ -1180,6 +1334,8 @@ void    qmgr_message_free(QMGR_MESSAGE *message)
 	myfree(message->client_name);
     if (message->client_addr)
 	myfree(message->client_addr);
+    if (message->client_port)
+	myfree(message->client_port);
     if (message->client_proto)
 	myfree(message->client_proto);
     if (message->client_helo)
@@ -1190,20 +1346,25 @@ void    qmgr_message_free(QMGR_MESSAGE *message)
 	myfree(message->sasl_username);
     if (message->sasl_sender)
 	myfree(message->sasl_sender);
+    if (message->log_ident)
+	myfree(message->log_ident);
     if (message->rewrite_context)
 	myfree(message->rewrite_context);
     recipient_list_free(&message->rcpt_list);
     qmgr_message_count--;
-    myfree((char *) message);
+    if ((message->tflags & DEL_REQ_FLAG_MTA_VRFY) != 0)
+	qmgr_vrfy_pend_count--;
+    myfree((void *) message);
 }
 
 /* qmgr_message_alloc - create in-core message structure */
 
 QMGR_MESSAGE *qmgr_message_alloc(const char *queue_name, const char *queue_id,
-				         int qflags)
+				         int qflags, mode_t mode)
 {
-    char   *myname = "qmgr_message_alloc";
+    const char *myname = "qmgr_message_alloc";
     QMGR_MESSAGE *message;
+    struct stat st;
 
     if (msg_verbose)
 	msg_info("%s: %s %s", myname, queue_name, queue_id);
@@ -1235,6 +1396,32 @@ QMGR_MESSAGE *qmgr_message_alloc(const char *queue_name, const char *queue_id,
 	return (0);
     } else {
 
+	/*
+	 * We have validated the queue file content, so it is safe to modify
+	 * the file properties now.
+	 */
+	if (mode != 0 && fchmod(vstream_fileno(message->fp), mode) < 0)
+	    msg_fatal("fchmod %s: %m", VSTREAM_PATH(message->fp));
+
+	/*
+	 * If this message is force-expired, use the existing defer logfile
+	 * records and do not assign any deliveries, leaving the refcount at
+	 * zero. If this message is force-expired, but no defer logfile
+	 * records are available, assign deliveries to the retry transport so
+	 * that the sender will still find out what recipients are affected
+	 * and why. Either way, do not assign normal deliveries because that
+	 * would be undesirable especially with mail that was expired in the
+	 * 'hold' queue.
+	 */
+	if ((message->qflags & QMGR_FORCE_EXPIRE) != 0
+	    && stat(mail_queue_path((VSTRING *) 0, MAIL_QUEUE_DEFER,
+				    queue_id), &st) == 0 && st.st_size > 0) {
+	    /* Use this defer log; don't assign deliveries (refcount ==  0). */
+	    message->flags = 1;			/* simplify downstream code */
+	    qmgr_message_close(message);
+	    return (message);
+	}
+
 	/*
 	 * Reset the defer log. This code should not be here, but we must
 	 * reset the defer log *after* acquiring the exclusive lock on the
@@ -1262,7 +1449,7 @@ QMGR_MESSAGE *qmgr_message_alloc(const char *queue_name, const char *queue_id,
 
 QMGR_MESSAGE *qmgr_message_realloc(QMGR_MESSAGE *message)
 {
-    char   *myname = "qmgr_message_realloc";
+    const char *myname = "qmgr_message_realloc";
 
     /*
      * Sanity checks.
